from typing import Optional
from datetime import datetime

from ..base_object import BaseDiscordObject
from .. import snowflake, user as ext_user


class VoiceState(BaseDiscordObject):
    '''VoiceState.'''

    guild_id: 'Optional[snowflake.Snowflake]' = None  # type: ignore # ? snowflake the guild id this voice state is for
    channel_id: 'Optional[snowflake.Snowflake]' = None  # type: ignore #  ?snowflake the channel id this user is connected to
    user_id: 'snowflake.Snowflake' = None  # type: ignore #  snowflake the user id this voice state is for
    member: 'Optional[ext_user.Member]' = None  # type: ignore # ? guild member object the guild member this voice state is for
    session_id: 'str' = None  # type: ignore #  string the session id for this voice state
    deaf: 'bool' = None  # type: ignore #  boolean whether this user is deafened by the server
    mute: 'bool' = None  # type: ignore #  boolean whether this user is muted by the server
    self_deaf: 'bool' = None  # type: ignore #  boolean whether this user is locally deafened
    self_mute: 'bool' = None  # type: ignore #  boolean whether this user is locally muted
    self_stream: 'Optional[bool]' = None  # type: ignore # ? boolean whether this user is streaming using "Go Live"
    self_video: 'bool' = None  # type: ignore #  boolean whether this user's camera is enabled
    suppress: 'bool' = None  # type: ignore #  boolean whether this user is muted by the current user
    request_to_speak_timestamp: 'Optional[datetime]' = None  # type: ignore #  ?ISO8601 timestamp the time at which the user requested to speak

    @property
    def _auto_map(self):
        return {
            'guild_id': snowflake.Snowflake,
            'channel_id': snowflake.Snowflake,
            'user_id': snowflake.Snowflake,
            'member': ext_user.Member,
            'session_id': str,
            'deaf': bool,
            'mute': bool,
            'self_deaf': bool,
            'self_mute': bool,
            'self_stream': bool,
            'self_video': bool,
            'suppress': bool,
            'request_to_speak_timestamp': datetime.fromtimestamp,
        }



from datetime import datetime
from typing import List, Optional
import copy

from ..base_object import BaseDiscordObject
from .. import snowflake, user as ext_user


class GuildMemberUpdate(BaseDiscordObject):
    '''GuildMemberUpdate.

    Attributes:
        roles ([Snowflake]): DEPRECATED: This will be replaced with a `Role` object soon.
        guild_id (Snowflake): The ID of the guild.
        user (User): A partial user object.
        nick (str): Nickname of the `user` in the guild.
        avatar (str): Members guild avatar hash.
        joined_at (datetime): When the user joined the guild.
        premium_since (datetime): When the user started boosting the guild.
        deaf (bool): Whether the user is deafened in voice channels.
        mute (bool): Whether the user is muted in voice channels.
        pending (bool): Whether the user has not yet passed the guild's Membership Screening requirements.
    '''

    guild_id: 'snowflake.Snowflake' = None  # type: ignore
    roles: 'List[snowflake.Snowflake]' = None  # type: ignore
    user: 'ext_user.User' = None  # type: ignore
    nick: 'Optional[str]' = None  # type: ignore
    avatar: 'Optional[str]' = None  # type: ignore
    joined_at: 'Optional[datetime]' = None  # type: ignore
    premium_since: 'Optional[datetime]' = None  # type: ignore
    deaf: 'Optional[bool]' = None  # type: ignore
    mute: 'Optional[bool]' = None  # type: ignore
    pending: 'Optional[bool]' = None  # type: ignore
    is_pending: 'Optional[bool]' = None  # type: ignore

    @property
    def _auto_map(self):
        return {
            'guild_id': snowflake.Snowflake,
            'roles': [snowflake.Snowflake],
            'user': ext_user.User,
            'nick': str,
            'avatar': str,
            'joined_at': datetime.fromisoformat,
            'premium_since': datetime.fromisoformat,
            'deaf': bool,
            'mute': bool,
            'pending': bool,
            'hoisted_role': copy.copy,
            'is_pending': bool,
        }



'''Collection of events which are less likely to be directly implemented by the end user.'''

from .typing_start import TypingStart
from .guild_member_update import GuildMemberUpdate
from .voice_state import VoiceState


__all__ = [
    'GuildMemberUpdate',
    'TypingStart',
    'VoiceState'
]



from datetime import datetime
from abc import ABC
from typing import Any, Dict, Optional
import warnings
from ..utilities import log


class BaseDiscordObject(ABC):
    '''Abstract base of all common discord objects. All subclasses map directly to an actual API object.'''

    _log = log.Log()
    _auto_map: dict = None  # type: ignore

    def __init__(self, data: Optional[dict] = None):
        '''Base generic __init__ function.

        Arguments:
            data (dict): Discord compliant dict for the given object.
        '''
        if type(self) is BaseDiscordObject:
            raise NotImplementedError('You cannot instance BaseDiscordObject directly. Use a child object.')

        if data is not None and isinstance(data, dict):
            self.from_dict(data)

    def from_dict(self, data: dict) -> 'BaseDiscordObject':
        '''Parse an object from a dictionary and return it.'''
        self._auto_dict(data)
        return self

    def to_dict(self) -> dict:
        '''Convert object to dictionary suitable for API or other generic useage.'''
        raise NotImplementedError(f'{self.__class__.__name__} does not yet implement this function.')

    def _auto_dict(self, data: 'Dict[str, Any]') -> 'BaseDiscordObject':
        '''Attempt an automatic conversion of a dict to this objects type.'''
        if self._auto_map is None:
            raise NotImplementedError(f'_auto_map not defined for {self.__class__}.')

        for attribute_key, value in data.items():
            self._log.debug(attribute_key)
            if attribute_key not in self._auto_map:
                warnings.warn(RuntimeWarning(f'Saw unexpected attribute key [{attribute_key}] in [{self.__class__}].'))
                setattr(self, attribute_key, value)
                continue
            function = self._auto_map[attribute_key]
            if (function is None) or (value is None):
                continue
            if isinstance(function, list):
                if len(function) != 1:
                    raise IndexError('Auto map has function list in excess of 1.')
                if not isinstance(value, list):
                    raise TypeError(f'Data from input dict [{attribute_key}:{value}] was not a list with list auto_mapped type [{function}].')
                setattr(self, attribute_key, list())
                target_list: list = getattr(self, attribute_key)
                for sub_value in value:
                    target_list.append(function[0](sub_value))
            else:
                setattr(self, attribute_key, function(value))
        return self

    def validate(self):
        '''Check for valid object type.

        Raises:
            NotImplementedError: Always.
        '''
        raise NotImplementedError('This object does not support validation.')

    @staticmethod
    def _fromtimestamp_milliseconds(time_stamp):
        return datetime.fromtimestamp(time_stamp / 1000)



from typing import Optional, List
from .base_object import BaseDiscordObject
from . import user as ext_user, snowflake, activity as ext_activity


class Presence(BaseDiscordObject):
    '''Presence.'''

    user: 'ext_user.User' = None  # type: ignore # user object the user presence is being updated for
    guild_id: 'snowflake.Snowflake' = None  # type: ignore # snowflake id of the guild
    status: 'str' = None  # type: ignore # string either "idle", "dnd", "online", or "offline"
    activities: 'List[ext_activity.Activity]' = None  # type: ignore # array of activity objects user's current activities
    client_status: 'ClientStatus' = None  # type: ignore # client_status object user's platform-dependent status

    @property
    def _auto_map(self):
        return {
            'user': ext_user.User,
            'guild_id': snowflake.Snowflake,
            'status': str,
            'activities': [ext_activity.Activity],
            'client_status': ClientStatus,
        }


class ClientStatus(BaseDiscordObject):
    '''ClientStatus.'''

    desktop: 'Optional[str]' = None  # tpye: ignore # string the user's status set for an active desktop (Windows, Linux, Mac) application session
    mobile: 'Optional[str]' = None  # tpye: ignore # string the user's status set for an active mobile (iOS, Android) application session
    web: 'Optional[str]' = None  # tpye: ignore # string the user's status set for an active web (browser, bot account) application session

    @property
    def _auto_map(self):
        return {
            'desktop': str,
            'mobile': str,
            'web': str,
        }



from typing import Optional, List
from datetime import datetime

from .base_object import BaseDiscordObject
from . import snowflake, enumerations


class Activity(BaseDiscordObject):
    '''Activity.'''

    id: str = None  # type: ignore # UNDOCUMENTED General ID for the given activity
    name: str = None  # type: ignore # string the activity's name
    type: enumerations.ACTIVITY_TYPE = None  # type: ignore # integer activity type
    url: Optional[str] = None  # type: ignore #? ?string stream url, is validated when type is 1
    created_at: datetime = None  # type: ignore # integer unix timestamp (in milliseconds) of when the activity was added to the user's session
    timestamps: 'Optional[TimeStamps]' = None  # type: ignore #? timestamps object unix timestamps for start and/or end of the game
    application_id: 'Optional[snowflake.Snowflake]' = None  # type: ignore #? snowflake application id for the game
    details: Optional[str] = None  # type: ignore #? ?string what the player is currently doing
    state: 'Optional[str]' = None  # type: ignore #? ?string the user's current party status
    emoji: 'Optional[ActivityEmoji]' = None  # type: ignore #? ?emoji object the emoji used for a custom status
    party: 'Optional[Party]' = None  # type: ignore #? party object information for the current party of the player
    assets: 'Optional[Assets]' = None  # type: ignore #? assets object images for the presence and their hover texts
    secrets: 'Optional[Secrets]' = None  # type: ignore #? secrets object secrets for Rich Presence joining and spectating
    instance: 'Optional[bool]' = None  # type: ignore #? boolean whether or not the activity is an instanced game session
    flags: 'Optional[enumerations.ACTIVITY_FLAGS]' = None  # type: ignore #? integer activity flags ORd together, describes what the payload includes
    buttons: 'Optional[List[Buttons]]' = None  # type: ignore #? array of buttons the custom buttons shown in the Rich Presence (max 2)

    @property
    def _auto_map(self):
        return {
            'id': str,
            'name': str,
            'type': enumerations.ACTIVITY_TYPE,
            'url': str,
            'created_at': self._fromtimestamp_milliseconds,
            'timestamps': TimeStamps,
            'application_id': snowflake.Snowflake,
            'details': str,
            'state': str,
            'emoji': ActivityEmoji,
            'party': Party,
            'assets': Assets,
            'secrets': Secrets,
            'instance': bool,
            'flags': enumerations.ACTIVITY_FLAGS,
            'buttons': [Buttons],
        }


class TimeStamps(BaseDiscordObject):
    '''TimeStamps.'''

    start: 'Optional[datetime]' = None  # type: ignore
    end: 'Optional[datetime]' = None  # type: ignore

    @property
    def _auto_map(self):
        return {
            'start': self._fromtimestamp_milliseconds,
            'end': self._fromtimestamp_milliseconds,
        }


class Party(BaseDiscordObject):
    '''Party.'''

    id: 'str' = None  # type: ignore
    size: 'Optional[List[int]]' = None

    @property
    def _auto_map(self):
        return {
            'id': str,
            'size': [int],
        }


class Assets(BaseDiscordObject):
    '''Assets.'''

    large_image: 'Optional[str]' = None  # type: ignore # The id for a large asset of the activity, usually a snowflake
    large_text: 'Optional[str]' = None  # type: ignore # Text displayed when hovering over the large image of the activity
    small_image: 'Optional[str]' = None  # type: ignore # The id for a small asset of the activity, usually a snowflake
    small_text: 'Optional[str]' = None  # type: ignore # Text displayed when hovering over the small image of the activity

    @property
    def _auto_map(self):
        return {
            'large_image': str,
            'large_text': str,
            'small_image': str,
            'small_text': str,
        }


class Secrets(BaseDiscordObject):
    '''Secrets.'''

    join: 'Optional[str]' = None  # type: ignore # the secret for joining a party
    spectate: 'Optional[str]' = None  # type: ignore # the secret for spectating a game
    match: 'Optional[str]' = None  # type: ignore # the secret for a specific instanced match

    @property
    def _auto_map(self):
        return {
            'join': str,
            'spectate': str,
            'match': str,
        }


class Buttons(BaseDiscordObject):
    '''Buttons.'''

    label: 'Optional[str]' = None  # the text shown on the button (1-32 characters)
    url: 'Optional[str]' = None  # the url opened when clicking the button (1-512 characters)

    @property
    def _auto_map(self):
        return {
            'join': str,
            'spectate': str,
            'match': str,
        }


class ActivityEmoji(BaseDiscordObject):
    '''ActivityEmoji.'''

    name: 'str' = None  # type: ignore # string the name of the emoji
    id: 'snowflake.Snowflake' = None  # type: ignore # snowflake the id of the emoji
    animated: 'Optional[bool]' = None  # boolean whether this emoji is animated

    @property
    def _auto_map(self):
        return {
            'name': str,
            'id': snowflake.Snowflake,
            'animated': bool,
        }




import enum


class COMMAND_TYPE(enum.IntEnum):
    '''Type of command.

    Attributes:
        CHAT_INPUT (int): Slash commands; a text-based command that shows up when a user types.
        USER (int): A UI-based command that shows up when you right click or tap on a user.
        MESSAGE (int): A UI-based command that shows up when you right click or tap on a message.
    '''
    CHAT_INPUT = 1  # Slash commands; a text-based command that shows up when a user types
    USER = 2  # A UI-based command that shows up when you right click or tap on a user
    MESSAGE = 3  # A UI-based command that shows up when you right click or tap on a message


class COMMAND_OPTION(enum.IntEnum):
    '''Valid types or groupings of command options.'''
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4  # Any integer between -2^53 and 2^53
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7  # Includes all channel types + categories
    ROLE = 8
    MENTIONABLE = 9  # Includes users and roles
    NUMBER = 10  # Any double between -2^53 and 2^53


class COMPONENT_TYPES(enum.IntEnum):
    '''Types of Components.'''
    ACTION_ROW = 1
    BUTTON = 2
    SELECT_MENU = 3


class BUTTON_STYLES(enum.IntEnum):
    '''Styles of Button objects.'''
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


class INTERACTION_TYPES(enum.IntEnum):
    '''Types of interactions.

    Attributes:
        PING: Used by server to get online status of webhook based applications. Handled by Dyscord in background.
        APPLICATION_COMMAND: Command was a `/slash` style command, typed by the user.
        MESSAGE_COMPONENT: Command was an interaction directly with a `Message` or `User`, the client right clicked on them.
        APPLICATION_COMMAND_AUTOCOMPLETE: User it attempting to complete an interaction as is asking for suggestions.
    '''
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3
    APPLICATION_COMMAND_AUTOCOMPLETE = 4


class INTERACTION_RESPONSE_TYPES(enum.IntEnum):
    '''Types of interaction responses.

    Attributes:
        PONG: ACK a Ping.
        CHANNEL_MESSAGE_WITH_SOURCE: Respond to an interaction with a message.
        DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE: ACK an interaction and edit a response later, the user sees a loading state.
        DEFERRED_UPDATE_MESSAGE: Only for components (Not slash commands), ACK an interaction and edit the original message later; the user does not see a loading state.
        UPDATE_MESSAGE: Only for components (Not slash commands), edit the message the component was attached to.
        APPLICATION_COMMAND_AUTOCOMPLETE_RESULT: Respond to an autocomplete interaction with suggested choices.

    '''
    PONG = 1  # ACK a Ping
    CHANNEL_MESSAGE_WITH_SOURCE = 4  # respond to an interaction with a message
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5  # ACK an interaction and edit a response later, the user sees a loading state
    DEFERRED_UPDATE_MESSAGE = 6  # for components, ACK an interaction and edit the original message later; the user does not see a loading state
    UPDATE_MESSAGE = 7  # for components, edit the message the component was attached to
    APPLICATION_COMMAND_AUTOCOMPLETE_RESULT = 8  # respond to an autocomplete interaction with suggested choices


class INTERACTION_CALLBACK_FLAGS(enum.IntFlag):
    '''Flags for callback methods.'''
    NONE = 0
    EPHEMERAL = (1 << 6)  # only the user receiving the message can see it
